/*
 * alterator security setup
 * alterator-secsetup-macrosblock.cpp - main program
 */
#ifndef RPM_V413
#define RPM_V413 0
#endif

#ifndef LIBDIR
#define LIBDIR "/usr/lib"
#endif

#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
#include <algorithm>
#include <iterator>
#include <vector>

#include <pwd.h>
#include <grp.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>
#include <string.h>

#include <rpm/rpmlib.h>
#include <rpm/rpmdb.h>
#if RPM_V413
#include <rpm/rpmts.h>
#endif

struct Record {
    int enable;
    std::string filename;
};

struct RecordMove {
    std::string from;
    std::string to;
};

std::vector<Record> records_ = {
    // LibreOffice
    {1, LIBDIR"/LibreOffice/program/classes/js.jar"},
    {1, LIBDIR"/LibreOffice/program/classes/ScriptProviderForBeanShell.jar"},
    {1, LIBDIR"/LibreOffice/program/classes/ScriptProviderForJavaScript.jar"},
    {1, LIBDIR"/LibreOffice/program/classes/ScriptProviderForJava.jar"},
    {1, LIBDIR"/LibreOffice/program/pythonscript.py"},
    {1, LIBDIR"/LibreOffice/program/services/scriptproviderforbeanshell.rdb"},
    {1, LIBDIR"/LibreOffice/program/services/scriptproviderforjavascript.rdb"},
    {1, LIBDIR"/LibreOffice/program/services/scriptproviderforpython.rdb"},
    {1, LIBDIR"/LibreOffice/share/config/soffice.cfg/cui/ui/macroassigndialog.ui"},
    {1, LIBDIR"/LibreOffice/share/config/soffice.cfg/cui/ui/macroassignpage.ui"},
    {1, LIBDIR"/LibreOffice/share/config/soffice.cfg/cui/ui/macroselectordialog.ui"},
    {1, LIBDIR"/LibreOffice/share/config/soffice.cfg/modules/BasicIDE"},
    {1, LIBDIR"/LibreOffice/share/config/soffice.cfg/uui/ui/macrowarnmedium.ui"},
    {1, LIBDIR"/LibreOffice/share/config/soffice.cfg/xmlsec/ui/macrosecuritydialog.ui"},
    // LibreOffice-still
    {1, LIBDIR"/LibreOffice-still/program/classes/js.jar"},
    {1, LIBDIR"/LibreOffice-still/program/classes/ScriptProviderForBeanShell.jar"},
    {1, LIBDIR"/LibreOffice-still/program/classes/ScriptProviderForJavaScript.jar"},
    {1, LIBDIR"/LibreOffice-still/program/classes/ScriptProviderForJava.jar"},
    {1, LIBDIR"/LibreOffice-still/program/pythonscript.py"},
    {1, LIBDIR"/LibreOffice-still/program/services/scriptproviderforbeanshell.rdb"},
    {1, LIBDIR"/LibreOffice-still/program/services/scriptproviderforjavascript.rdb"},
    {1, LIBDIR"/LibreOffice-still/program/services/scriptproviderforpython.rdb"},
    {1, LIBDIR"/LibreOffice-still/share/config/soffice.cfg/cui/ui/macroassigndialog.ui"},
    {1, LIBDIR"/LibreOffice-still/share/config/soffice.cfg/cui/ui/macroassignpage.ui"},
    {1, LIBDIR"/LibreOffice-still/share/config/soffice.cfg/cui/ui/macroselectordialog.ui"},
    {1, LIBDIR"/LibreOffice-still/share/config/soffice.cfg/modules/BasicIDE"},
    {1, LIBDIR"/LibreOffice-still/share/config/soffice.cfg/uui/ui/macrowarnmedium.ui"},
    {1, LIBDIR"/LibreOffice-still/share/config/soffice.cfg/xmlsec/ui/macrosecuritydialog.ui"},
    // VLC
    {1, LIBDIR"/vlc/plugins/lua/liblua_plugin.so"}
};

std::vector<RecordMove> recordsmove_ = {
    // LibreOffice
    {LIBDIR"/LibreOffice/program/services/scriptproviderforbeanshell.rdb", LIBDIR"/LibreOffice/program/scriptproviderforbeanshell.rdb"},
    {LIBDIR"/LibreOffice/program/services/scriptproviderforjavascript.rdb", LIBDIR"/LibreOffice/program/scriptproviderforjavascript.rdb"},
    {LIBDIR"/LibreOffice/program/services/scriptproviderforpython.rdb", LIBDIR"/LibreOffice/program/scriptproviderforpython.rdb"},
    // LibreOffice-still
    {LIBDIR"/LibreOffice-still/program/services/scriptproviderforbeanshell.rdb", LIBDIR"/LibreOffice-still/program/scriptproviderforbeanshell.rdb"},
    {LIBDIR"/LibreOffice-still/program/services/scriptproviderforjavascript.rdb", LIBDIR"/LibreOffice-still/program/scriptproviderforjavascript.rdb"},
    {LIBDIR"/LibreOffice-still/program/services/scriptproviderforpython.rdb", LIBDIR"/LibreOffice-still/program/scriptproviderforpython.rdb"},
};

// change file attributes
void change(std::vector<Record> *records) {
    for(std::vector<Record>::iterator it = records->begin(); it != records->end(); ++it) {
        if(it->enable != 1) {
            continue;
        }
        chown(it->filename.c_str(), 0, 0);
        chmod(it->filename.c_str(), 0);
    }
}

// move file to new location
void move(std::vector<RecordMove> *records, int restore) {
    for(std::vector<RecordMove>::iterator it = records->begin(); it != records->end(); ++it) {
        if(restore) {
            std::rename(it->to.c_str(), it->from.c_str());
        } else {
            std::rename(it->from.c_str(), it->to.c_str());
        }
    }
}

// return index if file name is in hdr, otherwise return -1
int hasFilename(Header hdr, const char * filename) {
#if RPM_V413
    int rindex = -1;
    char ** basenames;
    char ** dirnames;
    uint32_t * dirindexes;
    rpmtd basenames_td = rpmtdNew();
    rpmtd dirnames_td = rpmtdNew();
    rpmtd dirindexes_td = rpmtdNew();
    headerGet(hdr, RPMTAG_BASENAMES, basenames_td, HEADERGET_MINMEM);
    headerGet(hdr, RPMTAG_DIRNAMES, dirnames_td, HEADERGET_MINMEM);
    headerGet(hdr, RPMTAG_DIRINDEXES, dirindexes_td, HEADERGET_MINMEM);
    basenames = (char **)basenames_td->data;
    dirnames = (char **)dirnames_td->data;
    dirindexes = (uint32_t *)dirindexes_td->data;
    std::string fname;
    for(int i = 0; i < rpmtdCount(basenames_td); i++) {
        fname.clear();
        fname.append(dirnames[dirindexes[i]]);
        fname.append(basenames[i]);
        if(strcmp(fname.c_str(), filename) == 0) {
            rindex = i;
            break;
        }

    }
    rpmtdFree(basenames_td);
    rpmtdFree(dirnames_td);
    rpmtdFree(dirindexes_td);
    return rindex;
#else
    const char ** filelist;
    int filecount;
    rpmBuildFileList(hdr, &filelist, &filecount);
    for(int i = 0; i < filecount; ++i, *filelist++) {
        if(strcmp(*filelist, filename) == 0) {
            return i;
        }
    }
    return -1;
#endif
}

// restore attributes from rpm database
void restore(std::vector<Record> *records) {
#if RPM_V413
    rpmts ts;
    rpmtd username_td;
    rpmtd groupname_td;
    rpmtd mode_td;
#else
    rpmdb db;
#endif
    Header hdr;
    rpmdbMatchIterator mi;
    char ** username = NULL;
    char ** groupname = NULL;
    const unsigned short * mode = NULL;
    struct passwd *pw;
    struct group  *gr;
    int index = -1;

    rpmReadConfigFiles(NULL, NULL);

#if RPM_V413
    ts = rpmtsCreate();
    rpmtsSetVSFlags(ts, (rpmVSFlags_e)-1);
    rpmtsSetRootDir(ts, NULL);
    if (rpmtsOpenDB(ts, O_RDONLY) != 0)
    {
        std::cerr << "cannot open rpm database" << std::endl;
        exit(1);
    }
    username_td = rpmtdNew();
    groupname_td = rpmtdNew();
    mode_td = rpmtdNew();
#else
    if (rpmdbOpen("", &db, O_RDONLY, 0644) != 0) {
        std::cerr << "cannot open rpm database" << std::endl;
        exit(1);
    }
#endif

    for(std::vector<Record>::iterator it = records->begin(); it != records->end(); ++it) {

#if RPM_V413
        mi = rpmtsInitIterator(ts, (rpmTag)RPMTAG_BASENAMES, it->filename.c_str(), 0);
#else
        mi = rpmdbInitIterator(db, RPMTAG_BASENAMES, it->filename.c_str(), 0);
#endif

        if(mi == NULL) {
            continue;
        }
        while ((hdr = rpmdbNextIterator(mi)) != NULL) {
            index = hasFilename(hdr, it->filename.c_str());
            if(index == -1) {
                continue;
            }
#if RPM_V413
            rpmtdReset(username_td);
            rpmtdReset(groupname_td);
            rpmtdReset(mode_td);
            if( headerGet(hdr, RPMTAG_FILEUSERNAME, username_td, HEADERGET_MINMEM) == 0 || 
                headerGet(hdr, RPMTAG_FILEGROUPNAME, groupname_td, HEADERGET_MINMEM) == 0 ||
                headerGet(hdr, RPMTAG_FILEMODES, mode_td, HEADERGET_MINMEM)  == 0) {
                std::cerr << "error reading header" << std::endl;
                continue;
            }
            username = (char **) username_td->data;
            groupname = (char **) groupname_td->data;
            mode = (unsigned short *) mode_td->data;
#else
            if( headerGetEntry(hdr, RPMTAG_FILEUSERNAME, NULL, (void **) &username, NULL) == 0 ||
                headerGetEntry(hdr, RPMTAG_FILEGROUPNAME, NULL, (void **) &groupname, NULL) == 0 ||
                headerGetEntry(hdr, RPMTAG_FILEMODES, NULL, (void **) &mode, NULL) == 0) {
                std::cerr << "error reading header" << std::endl;
                continue;
            }
                
#endif
            pw = getpwnam(username[index]);
            gr = getgrnam(groupname[index]);
            if (pw != 0 && gr != 0) {
                chown(it->filename.c_str(), pw->pw_uid, gr->gr_gid);
                chmod(it->filename.c_str(), mode[index]);
            } else {
                std::cerr << "error getting userid or groupid" << std::endl;
            }
        }
    }
    rpmdbFreeIterator(mi);
#if RPM_V413
    rpmtsFree(ts);
    rpmtdFree(username_td);
    rpmtdFree(groupname_td);
    rpmtdFree(mode_td);
#else
    rpmdbClose(db);
#endif
}

int main(int argc, char const *argv[]) {
    if(argc > 1) {
        std::string cmd = argv[1];
        if(cmd.compare("enable") == 0) {
            restore(&records_);
            move(&recordsmove_, 0);
            change(&records_);
        }

        if(cmd.compare("disable") == 0) {
            move(&recordsmove_, 1);
            restore(&records_);
        }
    } else {
        std::cout << "Usage: " << argv[0] << " <command>" << std::endl;
    }
    return 0;
}
